#version 330 core  

#define NR_POINT_LIGHTS 1

layout (location = 0) out vec4 fragColor;
layout (location = 1) out vec4 brightColor;

struct DirLight {
    vec3 direction;
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

struct PointLight {
	vec3 position;
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;

	float constant;
	float linear;
	float quadratic;
};

in vec2 fragTexCoords;
in vec3 fragNormal;
in vec3 fragWSPosition;
in vec4 fragLightSpacePosition;

//out vec4 color;

uniform vec3 viewPos;
uniform DirLight dirLight;
uniform PointLight pointLight[NR_POINT_LIGHTS];

uniform sampler2D texture_diffuse1;
uniform sampler2D shadowMap;
uniform samplerCube pointShadowMap;
uniform float farPlane;

// Function prototypes
mat3 calcDirLight(vec3 fragNormal, vec3 viewDir, DirLight dirLight, vec3 textureValue);
mat3 calcPointLight(vec3 fragNormal, vec3 fragWSPosition, vec3 viewDir, PointLight light, vec3 textureValue);
mat3 attenuatePointLight(mat3 ambDiffSpec, vec3 fragWSPosition, PointLight light);
float calcShadow(vec4 fragLightSpacePosition, vec3 fragNormal, vec3 lightDirection);
float calcPointShadow(vec3 fragWSPosition, vec3 lightPos, float farPlane, vec3 fragNormal);

//clipping -> make conditional?
in vec4 interpolatedVertexEye;

void main()
{    


//clipping
	vec4 clip_plane = vec4(0,1,0,0);
    float clipPos = dot (interpolatedVertexEye.xyz, clip_plane.xyz) + clip_plane.w;
    if (clipPos < 0.0) {
        discard;
    }
  
	
	vec3 viewDir = normalize(viewPos - fragWSPosition);
	vec3 fragNormal = normalize(fragNormal);
	vec3 textureValue = vec3(texture(texture_diffuse1, fragTexCoords));
	mat3 ambDiffSpec = calcDirLight(fragNormal, viewDir, dirLight, textureValue);
	float shadow = 1.0f - calcShadow(fragLightSpacePosition, fragNormal, dirLight.direction);
    vec3 dirLightResult = vec3(ambDiffSpec[0] + shadow * (ambDiffSpec[1] + ambDiffSpec[2]));
	

	vec3 pointLightResult = vec3(0.0f, 0.0f, 0.0f);
	for(int i = 0; i < NR_POINT_LIGHTS; i++) {
		ambDiffSpec = calcPointLight(fragNormal, fragWSPosition, viewDir, pointLight[i], textureValue);
		ambDiffSpec = attenuatePointLight(ambDiffSpec, fragWSPosition, pointLight[i]);
		shadow = 1.0f - calcPointShadow(fragWSPosition, pointLight[i].position, farPlane, fragNormal);
		//pointLightResult = vec3(ambDiffSpec[0] + ambDiffSpec[1] + ambDiffSpec[2]);
		pointLightResult = vec3(ambDiffSpec[0] + shadow * (ambDiffSpec[1] + ambDiffSpec[2]));
	}

	//color = vec4(vec3((fragNormal.xyz + 1)/2), 1.0f);
	//vec4 col = texture(texture_diffuse1, fragTexCoords);

	// tone mapping
	//vec3 hdrColor = dirLightResult + pointLightResult;
	//const float gamma = 2.2;
	//vec3 mapped = hdrColor / (hdrColor + vec3(1.0));
	// Gamma correction 
    //mapped = pow(mapped, vec3(1.0 / gamma));
	//color = vec4(mapped, col.a);


	fragColor = vec4(dirLightResult + pointLightResult, 1.0);
	float brightness = dot(fragColor.rgb, vec3(0.2126, 0.7152, 0.0722)); // definition luminance
	if (brightness > 5) {
		brightColor = fragColor;
	}
	

	//color = vec4(pointLightResult, 1.0f);
	//color = vec4(vec3(shadow), 1.0f);
	//color = vec4(normalize(fragWSPosition - pointLight[0].position), 1.0f);
	//color = vec4(normalize(vec3(shadow)), 1.0f);
	//color = vec4(pointLightResult, 1.0f);

	//color = vec4(texture(shadowMap, fragTexCoords).r, 1.0f, 1.0f, 1.0f);
	//color = vec4(texture(texture_diffuse1, fragTexCoords));
	//color = vec4(specular, 1.0f);
	//color = vec4(max(dot(fragNormal, lightDir), 0.0), 1.0f, 1.0f, 1.0f);
}



mat3 calcDirLight(vec3 fragNormal, vec3 viewDir, DirLight dirLight, vec3 textureValue)
{
	vec3 lightDir = normalize(-dirLight.direction);
	vec3 halfwayDir = normalize(lightDir + viewDir);   //blinn phong
	// vec3 reflectDir = reflect(-lightDir, fragNormal); // phong

    float diff = max(dot(fragNormal, lightDir), 0.0);
    float spec = pow(max(dot(viewDir, halfwayDir), 0.0), 32);

	vec3 ambient = dirLight.ambient * textureValue;
	vec3 diffuse = dirLight.diffuse * diff * textureValue;
	vec3 specular = dirLight.specular * spec * textureValue;
    
	return mat3(ambient, diffuse, specular);
}

mat3 calcPointLight(vec3 fragNormal, vec3 fragWSPosition,
 vec3 viewDir, PointLight light, vec3 textureValue)
{
	vec3 lightDir = normalize(light.position - fragWSPosition);
	vec3 halfwayDir = normalize(lightDir + viewDir);   //blinn phong
	// vec3 reflectDir = reflect(-lightDir, fragNormal); // phong

    float diff = max(dot(fragNormal, lightDir), 0.0);
    float spec = pow(max(dot(viewDir, halfwayDir), 0.0), 32);

	vec3 ambient = light.ambient * textureValue;
	vec3 diffuse = light.diffuse * diff * textureValue;
	vec3 specular = light.specular * spec * textureValue;
    
	return mat3(ambient, diffuse, specular);
}

mat3 attenuatePointLight(mat3 ambDiffSpec, vec3 fragWSPosition, PointLight light) {
	
	// Attenuation
    float dist = length(light.position - fragWSPosition);
    float attenuation = 1.0f / (light.constant + light.linear * dist + light.quadratic * (dist * dist));  
	ambDiffSpec = attenuation * ambDiffSpec;
	return ambDiffSpec;
}

float calcShadow(vec4 fragLightSpacePosition, vec3 fragNormal, vec3 lightDirection) {
	// transform the light-space fragment position in clip-space to normalized device coordinates
	// (perpective divide)
	vec3 ndCoords = fragLightSpacePosition.xyz / fragLightSpacePosition.w;

	//  transform coordinates from [-1,1] to the range [0,1] 
	ndCoords = ndCoords * 0.5 + 0.5; 

	// Get closest depth value from light's perspective
   // float closestDepth = texture(shadowMap, ndCoords.xy).r; 

    // Get depth of current fragment from light's perspective
    float currentDepth = ndCoords.z;

	float shadow = 0.0f;
	// peter panning effect visible on the shadows of the fence if bias too high!
	float bias = max(0.00000005 * (1.0 - dot(fragNormal, lightDirection)), 0.000000005); 
	
	float closestDepth = 0.0f;

	// PCF (percentage closer filtering)
	// second argument of textureSize is lod (level)
	vec2 texelSize = 1.0 / textureSize(shadowMap, 0);
	for(int x = -2; x <= 2; ++x)
	{
		for(int y = -2; y <= 2; ++y)
		{
			closestDepth = texture(shadowMap, ndCoords.xy + vec2(x, y) * texelSize).r; 
			shadow += currentDepth - bias > closestDepth ? 1.0 : 0.0;        
		}    
	}
	shadow /= 25.0;

	// without PCF:
    // Check whether current frag pos is in shadow
    //shadow = currentDepth - bias > closestDepth  ? 1.0 : 0.0;

	 return shadow;

}


// array of offset direction for sampling for point shadows
vec3 sampleOffsetDirections[20] = vec3[]
(
   vec3(1, 1, 1), vec3(1, -1, 1), vec3(-1, -1, 1), vec3(-1, 1, 1), 
   vec3(1, 1, -1), vec3(1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1, 0), vec3(1, -1, 0), vec3(-1, -1, 0), vec3(-1, 1, 0),
   vec3(1, 0, 1), vec3(-1, 0, 1), vec3(1, 0, -1), vec3(-1, 0, -1),
   vec3(0, 1, 1), vec3(0, -1, 1), vec3(0, -1, -1), vec3(0, 1, -1)
);

float calcPointShadow(vec3 fragWSPosition, vec3 lightPos, float farPlane, vec3 fragNormal) {
	vec3 fragToLight = fragWSPosition - lightPos;

	// Get closest depth value from light's perspective
   //float closestDepth = texture(pointShadowMap, fragToLight).r; 
	// closestDepth value is currently in the range [0,1] so we first transform it back to [0,far_plane] 
	//closestDepth = closestDepth * farPlane;  

    float currentDepth = length(fragToLight);


	//float bias = 0.05;
	float bias = max(0.06 * (1.0 - dot(fragNormal, fragToLight)), 0.005); 
	float shadow = 0.0f;
	float samples = 20;
	float diskRadius = 0.05;
	float closestDepth = 0.0f;

	for(int i = 0; i < samples; ++i) {
		closestDepth = texture(pointShadowMap, fragToLight + sampleOffsetDirections[i] * diskRadius).r;
		closestDepth = closestDepth * farPlane;   // Undo mapping [0;1]
		if(currentDepth - bias > closestDepth) {
			shadow = shadow + 1.0;
		}
	}
	shadow = shadow / float(samples);  

    // Check whether current frag pos is in shadow
    //shadow = currentDepth - bias > closestDepth  ? 1.0 : 0.0;
	//float shadow = currentDepth > closestDepth  ? 1.0 : 0.0;

	 return shadow;

}


